////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

// ### bicg needs possible refining step

using namespace std;

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <complex>
#include "common.hpp"
#include "dyna.hpp"
#include "timer.hpp"
#include "info.hpp"
#include "element.hpp"
#include "vector.hpp"
#include "store.hpp"
#include "avl.hpp"
#include "sparse-nocomplex.hpp"
#include "hydra-steady.hpp"

void SteadyStateAnalyser::bicg(Vector <double> &solution, long maxIterations, double accuracy) {

/* BiConjugate Gradient from H.A. Van der Vorst - converges erratically */

  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting BiCG solve..." << endl;
  #endif

  long n;

  Vector<double> r, rTilde, p, pTilde, v, vTilde, result;
  double rho0, rho1, alpha, beta, solNorm;
  long iterations;

#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif

  result.setLength(tangible);
  r.setLength(tangible);
  rTilde.setLength(tangible);
  p.setLength(tangible);
  pTilde.setLength(tangible);
  v.setLength(tangible);
  vTilde.setLength(tangible);

  QRC->calcResidual(solution, r);
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    QRC->calcResidual(normSol, normRes);
    cout << "iteration " << 0 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << 0 << " " << log10(normRes.twoNorm()) << endl;  
#endif
  QRC->multiply(solution, rTilde);
  for (n = 0; n < tangible; n++) {
//    rTilde(n) = r(n); p[n] = r(n); pTilde[n] = r(n);
    rTilde[n] = -rTilde(n); p[n] = 0; pTilde[n] = 0;
//    rTilde[n] = r(n); p[n] = 0; pTilde[n] = 0;
//    rTilde[n] = -rTilde(n); p[n] = r(n); pTilde[n] = -rTilde(n); 
  }

  timer->read();
  if (solRepInterval && solRepStyle != NONE) {
    cout << "Iteration        Rho               Resid.-norm         Accuracy" << endl;
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
  }

  double old, diff, maxDiff = 1.0;
  iterations = 0;
  rho0 = (double) 1.0;

  /*
    To avoid an uninitialised warning is it actually possible that 
    the while loop condition fails the first time around?
   */
  rho1 = (double) rTilde.dot(r);
  while ((fabs(rho0) > EPS*0.001) && (maxDiff > accuracy) && iterations < maxIterations)  {
    rho1 = (double) rTilde.dot(r);
    beta = (double) rho1/rho0;
    for (n=0; n<tangible; n++) {
      p[n] = r(n) + beta*p(n);
    }
    QRC->transMultiply(p, v);
//    cout << "pTilde.dot(v) = " << pTilde.dot(v) << endl;
    // moveup
    for (n=0; n<tangible; n++) {
      pTilde[n] = rTilde(n) + beta*pTilde(n);
    }
    alpha = (double) rho1/(pTilde.dot(v));
    QRC->multiply(pTilde, vTilde);
 
    maxDiff = 0;
    solNorm = 0;
    for (n=0; n<tangible; n++) {
      old = solution(n);
      solution[n] += alpha*p(n);
      if (solution(n) > solNorm) solNorm = solution(n);
      diff = fabs(old - solution(n));
      if (diff > maxDiff) maxDiff = diff;
      r[n] -= (double) alpha*v(n);
    }
    // ### move up
//    cout << "r.dot(rTilde) = " << r.dot(rTilde) << endl;
    for (n=0; n<tangible; n++) {
      rTilde[n] -= (double) alpha*vTilde(n);
    }
    maxDiff /= solNorm;
    iterations++;
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    QRC->transMultiply(normSol, normRes);
    cout << "iteration " << iterations << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << iterations*2 << " " << log10(normRes.twoNorm()) << endl;
#endif

    if (solRepInterval && !(iterations % solRepInterval) && solRepStyle != NONE) {
      cout << setw(15) << iterations << " | " << setw(15) << rho1 << " | " << setw(17) << r.infNorm() << " | " << setw(17) << maxDiff << endl;
    }

#ifdef DEBUG
    QRC->transMultiply(solution, result);
    infnorm = result.infNorm();
    cout << "iteration " << iterations << ": infnorm is " << infnorm << "(" << r.infNorm() << ") (rho is " << rho1 << ")" << endl;
#endif
    rho0 = rho1;
  }

  if (solRepInterval && (iterations % solRepInterval) && solRepStyle != NONE) {
    cout << setw(15) << iterations << " | " << setw(15) << rho1 << " | " << setw(17) << r.infNorm() << " | " << setw(17) << maxDiff << endl;
  }
  if (solRepInterval && solRepStyle != NONE)
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
  if (maxDiff > accuracy && fabs(rho1) > EPS*0.001)
    terminate("BiCG failed to converge within %d iterations", maxIterations);

  cout.unsetf(ios::showpoint);
  if (maxDiff > accuracy) {
    cout << "\nwarning: BiCG could not achieve the desired accuracy (rho too small)." << endl;
    cout <<"         desired accuracy was " << accuracy << "; achieved accuracy was " << maxDiff << "." << endl << endl;
  }

  solution.normalise();
  Q->calcResidual(solution, result);
  checkViolation(checkVector(solution, result.infNorm()));
  unScale(solution);
  solution.normalise();
#ifndef RELEASE
  solution.textOutput("bicg.sol");
#endif
  solution.output(steadyVectorFile);  
}

void SteadyStateAnalyser::bicgstab(Vector<double> &solution, long maxIterations, double accuracy)
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting BiCGSTAB solve..." << endl;
  #endif

  long n;

  Vector<double> r, rTilde, v, p, s, t;
  register double omega, alpha, beta, beta2, rho0, rho1;
  double solNorm;
  long iterations;
#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif

  r.setLength(tangible);
  v.setLength(tangible);
  p.setLength(tangible);
  s.setLength(tangible);  
  t.setLength(tangible);  
  rTilde.setLength(tangible);  

  Q->calcResidual(solution,r);
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->calcResidual(normSol, normRes);
    cout << "iteration " << 0 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << 0 << " " << log10(normRes.twoNorm()) << endl;  
#endif

  for (n=0; n<tangible; n++)  {
#ifdef SRAND48
    rTilde[n] = 1000*drand48(); // r(n); // 20*drand48();//r(n);
#else
    rTilde[n] = 1000*rand()/32767.0;
#endif
    v[n] = 0;
    p[n] = 0;
  }

  timer->read();
  if (solRepInterval && solRepStyle != NONE) {
    cout << "Iteration        Rho               Resid.-norm         Accuracy" << endl;
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
  }
  rho0 = alpha = omega = 1.0;
  double old, diff, maxDiff = 1.0;
  iterations = 0;

  /*
    To avoid an uninitialised warning is it actually possible that 
    the while loop condition fails the first time around?
   */
  rho1 = rTilde.dot(r);

  while ((fabs(rho0) > EPS) && (maxDiff > accuracy) && iterations < maxIterations)  {
    rho1 = rTilde.dot(r);
//    cout << "rho1: " << rho1 << endl;
    beta = (double) (rho1*alpha)/(rho0*omega);
    beta2 = (double) (rho1*alpha)/rho0;
    for (n=0; n<tangible; n++)
      p[n] = beta*p(n) - beta2*v(n) + r(n);
    Q->transMultiply(p, v);
    alpha = (double) rho1 / (double) rTilde.dot(v);
    for (n=0; n<tangible; n++)
      s[n] = r(n) - alpha*v(n);
    Q->transMultiply(s, t);
    omega = (double) t.dot(s) / (double) t.dot(t);
    maxDiff = 0;
    solNorm = 0;
    for (n=0; n<tangible; n++) {
      old = solution(n);
      solution[n] += (alpha*p(n) + omega*s(n));
      if (solution(n) > solNorm) solNorm = solution(n);
      diff = fabs(old - solution(n));
      if (diff > maxDiff) maxDiff = diff;
      r[n] = s(n) - omega*t(n);
    }
    maxDiff /= solNorm;
    iterations++;
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    cout << "iteration " << iterations << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << iterations*2 << " " << log10(normRes.twoNorm()) << endl;
#endif

    if (solRepInterval && !(iterations % solRepInterval) && solRepStyle != NONE) {
      cout << setw(15) << iterations << " | " << setw(15) << rho1 << " | " << setw(17) << r.infNorm() << " | " << setw(17) << maxDiff << endl;
    }
    rho0 = rho1;
  }

  if (solRepInterval && (iterations % solRepInterval) && solRepStyle != NONE) {
    cout << setw(15) << iterations << " | " << setw(15) << rho1 << " | " << setw(17) << r.infNorm() << " | " << setw(17) << maxDiff << endl;
  }
  if (solRepInterval && solRepStyle != NONE)
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
  if (maxDiff > accuracy && fabs(rho1) > EPS)
    terminate("BiCGSTAB failed to converge within %d iterations", maxIterations);

  cout.unsetf(ios::showpoint);

  int acc = 0;
  if (maxDiff > accuracy) {
    cout << "\nwarning: BiCGSTAB could not achieve the desired accuracy (rho too small)." << endl;
    cout <<"         desired accuracy was " << accuracy << "; achieved accuracy was " << maxDiff << "." << endl << endl;
    acc = 1;
  }

  solution.normalise();
  Q->calcResidual(solution, p);

  long status = checkVector(solution, p.infNorm());
  double timeUsed = timer->read();
  if (status & RANGE_VIOLATION || acc) {
    refine(solution, maxIterations, accuracy);
    timeUsed += timer->read();
  } else 
    checkViolation(status);
  unScale(solution);
  solution.normalise();

  solution.output(steadyVectorFile);  
  #ifndef RELEASE
  solution.textOutput("cgstab.sol");
  #endif
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): BiCGSTAB complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}


void SteadyStateAnalyser::bicgstab2(Vector <double> &solution, long maxIterations, double accuracy) {
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting BICGSTAB(2) solve..." << endl;
#endif
  long n;
  Vector<double> ri, rTilde, r, s, t, u, v, w, x, resid;
  double rho0, rho1, alpha, beta, gamma, omega1, omega2, mu, nu, tau;
  register double factor, maxDiff, convergence, solNorm;
#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif

  ri.setLength(tangible);
  r.setLength(tangible);
  s.setLength(tangible);
  t.setLength(tangible);
  rTilde.setLength(tangible);
  u.setLength(tangible);
  x.setLength(tangible);
  v.setLength(tangible);
  w.setLength(tangible);
  // debug only
  resid.setLength(tangible);

  Q->calcResidual(solution, ri);

#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->calcResidual(normSol, normRes);
    cout << "iteration " << 0 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << 0 << " " << log10(normRes.twoNorm()) << endl;  
#endif

  for (n=0; n<tangible; n++) {
    u[n] = 0;
    rTilde[n] = ri(n);
  }

  rho0 = 1;
  alpha = 0;
  omega2 = 1;

  timer->read();
  if (solRepInterval && solRepStyle != NONE) {
    cout << "Iteration        Rho               Resid.-norm bound   Accuracy" << endl;
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
  }

  for (long iterations = 2; iterations < maxIterations; iterations+=2) {

    rho0 *= -omega2;

    // even bicg step
    rho1 = rTilde.dot(ri);
    beta = alpha*rho1/rho0;
    rho0 = rho1;

    for (n=0; n<tangible; n++) 
      u[n] = ri[n] - beta*u[n];
    Q->transMultiply(u,v);

    gamma = rTilde.dot(v);
    alpha = rho0/gamma;
    for (n=0; n<tangible; n++) 
      r[n] = ri[n] - alpha*v[n];
    Q->transMultiply(r,s);
    for (n=0; n<tangible; n++) 
      x[n] = solution[n] + alpha*u[n];

    // odd bicg step
    rho1 = rTilde.dot(s);
    beta = alpha*rho1/rho0;
    rho0 = rho1;
    for (n=0; n<tangible; n++)
      v[n] = s[n] - beta*v[n];
    Q->transMultiply(v,w);
    gamma = rTilde.dot(w);
    alpha = rho0/gamma;
    for (n=0; n<tangible; n++) {
      u[n] = r[n] - beta*u[n];
      r[n] = r[n] - alpha*v[n];
      s[n] = s[n] - alpha*w[n];
    }
    Q->transMultiply(s,t);

    // gcr(2) part
    omega1 = r.dot(s);
    mu = s.dot(s);
    nu = s.dot(t);
    tau = t.dot(t);
    omega2 = r.dot(t);
    tau = tau - nu*nu/mu;
    omega2 = (omega2 - nu*omega1/mu)/tau;
    omega1 = (omega1 - nu*omega2)/mu;
    maxDiff = 0;
    solNorm = 0;
    for (n=0; n<tangible; n++) {
      factor = omega1*r(n) + omega2*s(n) + alpha*u(n);
      if (fabs(factor) > maxDiff) maxDiff = fabs(factor);
      solution[n] = x(n) + factor;
      if (solution(n) > solNorm) solNorm = solution(n);
      ri[n] = r(n) - omega1*s(n) - omega2*t(n);
      u[n] = u[n] - omega1*v[n] - omega2*w[n];
    }
    convergence = maxDiff/solNorm;

    cout << setw(15) << iterations << " | " << setw(15) << rho0 << " | " << setw(17) << ri.infNorm() << " | " << setw(17) << convergence << endl;

#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    cout << "iteration " << iterations << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << iterations*2 << " " << log10(normRes.twoNorm()) << endl;
#endif
//    cout << "iteration: " << iterations << endl;
//    cout << "  ri.twoNorm = " << ri.twoNorm() << endl;
//    Q->calcResidual(solution, resid);
//    cout << "  true resid norm = " << resid.twoNorm() << endl;
//    cout << "  maxDiff = " << maxDiff << endl;
//    cout << "  convergence = " << convergence << endl;
    if (convergence < accuracy) break;
  }
  if (solRepInterval && solRepStyle != NONE)
    cout << "----------------+-----------------+-------------------+-------------------" << endl;

  cout.unsetf(ios::showpoint);

  solution.normalise();
  Q->calcResidual(solution, w);
  long status = checkVector(solution, w.infNorm());
  double timeUsed = timer->read();

  if (status & RANGE_VIOLATION) {
    refine(solution, maxIterations, accuracy);
    timeUsed += timer->read();
  } else 
    checkViolation(status);
  unScale(solution);
  solution.normalise();
  solution.output(steadyVectorFile);
  #ifndef RELEASE
  solution.textOutput("bicgstab2.sol");
  #endif
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): BICGSTAB(2) complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}

void SteadyStateAnalyser::cgs(Vector<double> &solution, long maxIterations, double accuracy)
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting CGS solve..." << endl;
  #endif
  // true residual version (seems to work better) 
  long n;

  Vector<double> r, rTilde, p, q, u, v, w, result;
  register double rho0, rho1, alpha, beta, bbeta;
  double solNorm;
  register double factor, maxDiff = 1.0;
#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif

  r.setLength(tangible);
  rTilde.setLength(tangible);
  p.setLength(tangible);
  q.setLength(tangible);  
  u.setLength(tangible);
  v.setLength(tangible);
  w.setLength(tangible);  
  result.setLength(tangible);  

  Q->calcResidual(solution,r);
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->calcResidual(normSol, normRes);
    cout << "iteration " << 0 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << 0 << " " << log10(normRes.twoNorm()) << endl;  
#endif
  for (n=0; n<tangible; n++)  {
    rTilde(n) = r(n);
    p(n) = 0;
    q(n) = 0;
  }

  rho0 = 1;
  long iterations = 0;

  timer->read();
  if (solRepInterval && solRepStyle != NONE) {
    cout << "Iteration        Rho               Resid.-norm         Accuracy" << endl;
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
  }

  /*
    To avoid an uninitialised warning is it actually possible that 
    the while loop condition fails the first time around?
   */
  rho1 = rTilde.dot(r);

  while ((fabs(rho0) > EPS) && (maxDiff > accuracy) && (iterations < maxIterations)) {
    rho1 = rTilde.dot(r);
    beta = rho1/rho0;
    bbeta = beta*beta;
    for (n=0; n<tangible; n++) {
      q[n] *= beta;
      u[n] = r(n) + q(n);
      p[n] = u(n) + q(n) + bbeta*p(n);
    }
    Q->transMultiply(p, v);
    alpha = rho1/rTilde.dot(v);
    maxDiff = 0;
    solNorm = 0;
    for (n=0; n<tangible; n++) {
      q(n) = u(n) - alpha*v(n);
      w(n) = u(n) + q(n);
      factor = alpha*w(n);
      solution(n) += factor;
      if (solution(n) > solNorm) solNorm = solution(n);
      if (fabs(factor) > maxDiff) maxDiff = fabs(factor);
    }
    maxDiff /= solNorm;

    Q->calcResidual(solution, r);
#ifdef DEBUG
    double infnorm = result.infNorm();
    cout << "iteration " << iterations << ": inf is " << infnorm << " (rho1 = " << rho1 << ", rinf is " << r.infNorm() << ") diff = " << diff << endl;
#endif
//    for (n=0; n<tangible; n++)
//      r(n) = -result(n);
    iterations++;
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    cout << "iteration " << iterations << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << iterations*2 << " " << log10(normRes.twoNorm()) << endl;
#endif
    if (solRepInterval && !(iterations % solRepInterval) && solRepStyle != NONE) {
      cout << setw(15) << iterations << " | " << setw(15) << rho1 << " | " << setw(17) << r.infNorm() << " | " << setw(17) << maxDiff << endl;
    }
    rho0 = rho1;
  }
  
  if (solRepInterval && (iterations % solRepInterval) && solRepStyle != NONE) {
    cout << setw(15) << iterations << " | " << setw(15) << rho1 << " | " << setw(17) << r.infNorm() << " | " << setw(17) << maxDiff << endl;
  }
  if (solRepInterval && solRepStyle != NONE)
    cout << "----------------+-----------------+-------------------+-------------------" << endl;

  if (maxDiff > accuracy && fabs(rho1) > EPS)
    terminate("CGS failed to converge within %d iterations", maxIterations);

  cout.unsetf(ios::showpoint);

  int acc = 0;
  if (maxDiff > accuracy) {
    cout << "\nwarning: CGS could not achieve the desired accuracy (rho too small)." << endl;
    cout <<"         desired accuracy was " << accuracy << "; achieved accuracy was " << maxDiff << "." << endl << endl;
    acc = 1;
  }

  solution.normalise();
  Q->calcResidual(solution, result);
  long status = checkVector(solution, result.infNorm());
  double timeUsed = timer->read();
  if (status & RANGE_VIOLATION || acc) {
    refine(solution, maxIterations, accuracy);
    timeUsed += timer->read();
  } else 
    checkViolation(status);
  unScale(solution);
  solution.normalise();

  solution.output(steadyVectorFile);  
  #ifndef RELEASE
  solution.textOutput("cgs.sol");
  #endif
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): CGS complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}

void SteadyStateAnalyser::cgnr(Vector<double> &solution, long maxIterations, double accuracy)
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting CGNR solve..." << endl;
  #endif

  long n;

  Vector<double> r, v, p, pTilde;
  double rho1, rho2, alpha, beta, solNorm;
  long iterations;
#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif

  r.setLength(tangible);
  v.setLength(tangible);
  p.setLength(tangible);  
  pTilde.setLength(tangible);  

  timer->read();
  QRC->calcResidual(solution, r);
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    QRC->calcResidual(normSol, normRes);
    cout << "iteration " << 0 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << 0 << " " << log10(normRes.twoNorm()) << endl;  
#endif
  for (n=0; n<tangible; n++)
    v(n) = 0;

  beta = 0;
  iterations = 0;
  double old, diff, maxDiff = 1;
  rho1 = r.dot(r);

  if (solRepInterval && solRepStyle != NONE) {
    cout << "Iteration        Rho               Resid.-norm         Accuracy" << endl;
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
  }

  
  /*
    To avoid an uninitialised warning is it actually possible that 
    the while loop condition fails the first time around?
   */
  rho2 = r.dot(r);

  while ((fabs(rho1) > EPS) && (maxDiff > accuracy) && iterations < maxIterations)  {
    QRC->multiply(r, pTilde);
    for (n=0; n<tangible; n++)
      v(n) = pTilde(n) + beta*v(n);
    alpha = rho1/v.dot(v);
    QRC->transMultiply(v, p);
    maxDiff = 0;
    solNorm = 0;
    for (n=0; n<tangible; n++) {
      old = solution(n);
      solution(n) += alpha*v(n);
      if (solution(n) > solNorm) solNorm = solution(n);
      diff = fabs(old - solution(n));
      if (diff > maxDiff) maxDiff = diff;
      r(n) -= alpha*p(n);
    }
    maxDiff /= solNorm;
    rho2 = r.dot(r);
    beta = rho2/rho1;
    iterations++;
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    QRC->transMultiply(normSol, normRes);
    cout << "iteration " << iterations << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << iterations*2 << " " << log10(normRes.twoNorm()) << endl;
#endif
    if (solRepInterval && !(iterations % solRepInterval) && solRepStyle != NONE) {
      cout << setw(15) << iterations << " | " << setw(15) << rho2 << " | " << setw(17) << r.infNorm() << " | " << setw(17) << maxDiff << endl;
    }
#ifdef DEBUG
    cout << "iteration " << iterations << ": rho = " << rho2 << ", rinf = " << r.infNorm() << endl;
#endif
    // check accuracy here
    rho1 = rho2;
  }
#ifdef DEBUG
  QRC->transMultiply(solution,p);
  cout << "result's infNorm: " << p.infNorm() << endl;
#endif
    if (solRepInterval && (iterations % solRepInterval) && solRepStyle != NONE) {
    cout << setw(15) << iterations << " | " << setw(15) << rho2 << " | " << setw(17) << r.infNorm() << " | " << setw(17) << maxDiff << endl;
  }

  if (solRepInterval && solRepStyle != NONE)
    cout << "----------------+-----------------+-------------------+-------------------" << endl;

  if (maxDiff > accuracy && fabs(rho1) > EPS)
    terminate("CGNR failed to converge within %d iterations", maxIterations);

  cout.unsetf(ios::showpoint);
  if (maxDiff > accuracy) {
    cout << "\nwarning: CGNR could not achieve the desired accuracy (rho too small)." << endl;
    cout <<"         desired accuracy was " << accuracy << "; achieved accuracy was " << maxDiff << "." << endl << endl;
  }

  solution.normalise();
  Q->calcResidual(solution, p);
  checkViolation(checkVector(solution, p.infNorm()));
  unScale(solution);
  solution.normalise();
  double timeUsed = timer->read();
  solution.output(steadyVectorFile);  
  #ifndef RELEASE
  solution.textOutput("cgnr.sol");
  #endif

  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): CGNR complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}

/*
void SteadyStateAnalyser::bicg()
{
  // siam templates - hopefully this will be better 
  long n, iterations;

//  cout << "AFTER MAKE Q: " << endl;
//  QRC->print();  

  Vector<double> solution, result, p, pTilde, q, qTilde, r, rTilde;
  double alpha, beta = 0, rho1, rho2, infnorm;

  solution.setLength(tangible);
  result.setLength(tangible);
  r.setLength(tangible);
  rTilde.setLength(tangible);
  p.setLength(tangible);
  pTilde.setLength(tangible);
  q.setLength(tangible);
  qTilde.setLength(tangible);

  for (n = 0; n < tangible; n++) solution[n] = 1;
  QRC->calcResidual(solution, r);
  for (n = 0; n < tangible; n++) {
    rTilde[n] = r(n); 
  }  
  iterations = 0;
  rho1 = rTilde.dot(r);
  while (fabs(rho1) > EPS) {
    cout << "### ITERATION " << iterations << endl;
    rho1 = rTilde.dot(r);
    if (iterations==0) {
      for (n=0; n<tangible; n++) {
        p[n] = r(n);
        pTilde[n] = rTilde(n);
      }
    } else {
      beta = rho1/rho2;
      for (n=0; n<tangible; n++) {
        p[n] = r(n) + beta*p(n);
        pTilde[n] = rTilde(n) + beta*pTilde(n);
      }
    }
    QRC->transMultiply(p, q);
    QRC->multiply(pTilde, qTilde);
    alpha = rho1 / pTilde.dot(q);
    for(n=0; n<tangible; n++) {
      solution[n] += alpha*p(n);
      r[n] -= alpha*q(n);
      rTilde[n] -= alpha*qTilde(n);
    }
    cout << "r: " << r.infNorm() << endl;
    cout << "rTilde: " << rTilde.infNorm() << endl;
    cout << "p: " << p.infNorm() << endl;
    cout << "pTilde: " << pTilde.infNorm() << endl;
    cout << "q: " << q.infNorm() << endl;
    cout << "qTilde: " << qTilde.infNorm() << endl;
    QRC->transMultiply(solution, result);
    infnorm = result.infNorm();
    cout << "iteration " << iterations << ": inf is " << infnorm << " (rho = " << rho1 << ", r.inf = " << r.infNorm() << ")" << endl;
    cout << "alpha = " << alpha << endl;
    cout << "beta = " << beta << endl;
    rho2 = rho1;
    iterations++;
  }

}
*/
/*
void SteadyStateAnalyser::bicg() {
  // stewart - converges really slowly - dead in the water
  long n;

  cout << "AFTER MAKE Q: " << endl;
  QRC->print();  

  Vector<double> solution, result, v, vTilde, p, pTilde, r, rTilde;
  double alpha, beta, rho1, rho2, infnorm;

  solution.setLength(tangible);
  result.setLength(tangible);
  r.setLength(tangible);
  rTilde.setLength(tangible);
  p.setLength(tangible);
  pTilde.setLength(tangible);
  v.setLength(tangible);
  vTilde.setLength(tangible);

  for (n = 0; n < tangible; n++) solution[n] = 1;
  QRC->calcResidual(solution, r);
  for (n = 0; n < tangible; n++) {
    cout << "r(" << n << ") = r~(" << n << ") = " << r(n) << endl;
    rTilde[n] = r(n); v[n] = r(n); vTilde[n] = r(n);
  }

  rho1 = rTilde.dot(r);  
  for (int i=0; i<100; i++) {
    cout << "### ITERATION: " << i << endl;
    QRC->transMultiply(v, p);
    QRC->multiply(v, pTilde);
    alpha = rho1/vTilde.dot(p);
    r.subtractScaled(alpha, p);
    rTilde.subtractScaled(alpha, pTilde);
    solution.addScaled(alpha, v);
//    cout << "current solution: " << solution << endl;
    QRC->transMultiply(solution, result);
    infnorm = result.infNorm();
    cout << "iteration " << i << ": infnorm is " << infnorm << "(rho = " << rho1 << ")" << endl;

    rho2 = rTilde.dot(r);
    beta = rho2/rho1;
    v.multiply(beta);
    v.add(r);
    vTilde.multiply(beta);
    vTilde.add(rTilde);
    rho1 = rho2;
  }
  unScale(solution);
  solution.normalise();
//  for (n=0; n<tangible; n++) {
//    cout << "final[" << n << "] = " << setprecision(20) << solution(n) << endl;
//  }
  cout << "result's infNorm: " << result.infNorm() << endl;
  #ifndef RELEASE
  solution.textOutput("cg.sol");
  #endif
  solution.output(steadyVectorFile);  
}*/

