////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <complex>
#include "common.hpp"
#include "dyna.hpp"
#include "timer.hpp"
#include "info.hpp"
#include "element.hpp"
#include "vector.hpp"
#include "store.hpp"
#include "avl.hpp"
#include "sparse-nocomplex.hpp"
#include "hydra-steady.hpp"

void SteadyStateAnalyser::grassmann(Vector <double> &solution)
// ### change to work on original, not QD!
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting Grassmann solve..." << endl;
  #endif

  long i,j,k,ri,rj, row, col;

#ifdef DEBUG
  long stage1=0, stage2=0;
#endif
  register SparseMatrixRow *_row;
  register SparseMatrixCol *_col;
  double first, second, sum;

  DirectMatrix *work = new DirectMatrix(tangible, tangible);
  for (i=0; i<tangible; i++) {
    _col = &QC->getCol(i);
    for (rj=0; rj<_col->getEntries(); rj++) {
      j = _col->getRow(rj);
      work->addElement(j, i, QC->getColElement(_col, rj));
    }
  }

  timer->read();
 
  for (k=tangible - 1; k>0; k--) {
    if (solRepInterval && !( (tangible - k + 1) % solRepInterval) && solRepStyle != NONE) cout << "eliminating row " << tangible - k + 1 << "..." << endl;
    sum = 0;
    _row = &work->getRow(k);
    _col = &work->getCol(k);
    for (rj=0; rj<_row->getEntries(); rj++) {
      col = _row->getColumn(rj);
      if (col < k) sum += work->getRowElement(_row, rj);
    }

    // ### DELETE THIS STUFF - this is GAUSS not GRASSMANN
    //    cout << "sum before is " << sum << endl;
    //    sum = -1.0/work->getElement(k,k);
    //    cout << "sum after is " << sum << endl;

    if (!sum)
      terminate("zero pivot sum encountered in grassmann");

    for (ri=0; ri<_col->getEntries(); ri++) {
      if (_col->getRow(ri) < k)
        work->divideColElement(_col, ri, sum);
    }

    for (ri = 0; ri<_col->getEntries(); ri++) {
      i = _col->getRow(ri);
#ifdef DEBUG
      stage1++;
#endif
      if (i < k) {
        first = work->getColElement(_col, ri);
	for (rj=0; rj<_row->getEntries(); rj++) {
	  j = _row->getColumn(rj);
#ifdef DEBUG
	  stage2++;
#endif
	  if (j < k && i != j) {
	    second = work->getRowElement(_row,rj);
            work->addToElement(i, j, first*second);
	  }
	}
      }
    }
  }
  solution[0] = 1;
  for (j=1; j<tangible; j++) {
    sum = 0;
    _col = &work->getCol(j);
    for (ri=0; ri<_col->getEntries(); ri++) {
      row = _col->getRow(ri);
      if (row < j)
        sum += work->getColElement(_col,ri)*solution(row);
    }    
    solution[j] = sum;
  }
  solution.normalise();
  double timeUsed = timer->read();
  cout << "solution complete." << endl << endl;
#ifdef DEBUG
  cout << "stage1 = " << stage1 << "; stage2 = " << stage2 << endl;
#endif

  Vector<double> result;
  result.setLength(tangible);
  QC->calcResidual(solution, result);
  long status = checkVector(solution, result.infNorm());
  if (status & RANGE_VIOLATION) {
//    refine(1.0, solution);
    QC->calcResidual(solution, result);
    status = checkVector(solution, result.infNorm());
  }
  checkViolation(status);
  delete work;
  solution.output(steadyVectorFile);
  #ifndef RELEASE
  solution.textOutput("grass.sol");
  #endif
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): Grassmann complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}

void SteadyStateAnalyser::gauss(Vector<double> &solution)
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting Gauss elimination..." << endl;
  #endif

  DirectMatrix *reduced = new DirectMatrix(tangible, tangible);

  register long i,j,k;
#ifdef DEBUG
  long stage1 = 0, stage2 = 0;
#endif
  long col;
  double pivot;
  double *_row = new double[tangible];
  double tol1 = EPS, tol2 = EPS/100;
  register SparseMatrixRow *_rowp;
  register SparseMatrixCol *_col;

  timer->read();

  for (k=0; k<tangible; k++) _row[k] = 0;

  for (k=0; k<tangible - 1; k++) {
    if (solRepInterval && !(k % solRepInterval) && solRepStyle != NONE) cout << "eliminating row " << k << "..." << endl;
    _col = &QC->getCol(k);
    for (j=0; j<_col->getEntries(); j++)
      _row[_col->getRow(j)] = QC->getColElement(_col, j);

    for (j=0; j<k; j++) { // eliminate first k entries in current row
      if (fabs(_row[j]) > tol1) {
#ifdef DEBUG
        stage1++;
#endif
        pivot = 1.0/_row[j];
        for (i = j; i<tangible; i++) _row[i] *= pivot;
        _rowp = &reduced->getRow(j);
        for (i = 0; i<_rowp->getEntries(); i++) {         
#ifdef DEBUG
          stage2++;
#endif
          col = _rowp->getColumn(i);
          if (col >= j)
            _row[col] -= reduced->getRowElement(_rowp,i);
	}
      }
    }

    if (!_row[k]) 
      terminate("zero pivot encountered in gauss elimination");
    pivot = 1.0/_row[k];    
    for (i=k; i<tangible; i++) {
      _row[i] *= pivot;
      if (fabs(_row[i]) > tol2)
        reduced->addElement(k, i, _row[i]);
      _row[i] = 0;
    }
  }

  double sum;
  solution[tangible - 1] = 1;
  for (k = tangible - 2; k >= 0; k--) {
    sum = 0;
    pivot = 0;
    _rowp = &reduced->getRow(k);
    for (j=0; j<_rowp->getEntries(); j++) {
      col = _rowp->getColumn(j);
//      assert(col >= k);
      if (col==k) pivot = reduced->getRowElement(_rowp,j);
        else sum += reduced->getRowElement(_rowp,j)*solution(col);
    }
    if (!pivot)
      terminate("zero pivot encountered during back-substitution");

    solution[k] = -sum/pivot;
  }

  delete _row;
//  unScale(solution);
  solution.normalise();
  double timeUsed = timer->read();
#ifdef DEBUG
  cout << "stage1 = " << stage1 << "; stage2 = " << stage2 << endl;
#endif
  delete reduced;

  cout << "solution complete." << endl << endl;
  Vector<double> result;
  result.setLength(tangible);
  QC->calcResidual(solution, result);
  long status = checkVector(solution, result.infNorm());
  if (status & RANGE_VIOLATION) {
    reScale(solution);
    refine(solution, maxIterations, accuracy);
  } else
    checkViolation(status);

  solution.output(steadyVectorFile);
  #ifndef RELEASE
  solution.textOutput("gauss.sol");
  #endif
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): Gauss elimination complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}


