////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

//#define DEBUG
//#define BUFFER_SIZE 16384

using namespace std;

#define USE_STEADY

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstdarg>
#include <complex>
#include <cmath>
#include "common.hpp"
#include "dyna.hpp"
#include "timer.hpp"
#include "info.hpp"
#include "element.hpp"
#include "vector.hpp"
#include "user.hpp"
#include "full.hpp"
#include "target.hpp"
#include "filenames.cxx"
#ifdef RESOURCE_USAGE
#include <unistd.h>
#endif

#ifndef LOGGING
#define SILENT 1
#endif

void TargetAnalyser::terminate(char *fmt, ...)
{
  char buffer[2048];

  va_list argp;
  va_start(argp, fmt);
  vsprintf(buffer, fmt, argp); 
  va_end(argp);
  destroy();
  printf("\nfatal error: %s.\nexecution halted.\n", buffer);  
  exit(1);
}

TargetAnalyser::TargetAnalyser(char* basename)
{
   char* optionsFileName = add_extension(basename, "OPTIONS") ;
   char* infoFileName    =  add_extension(basename, "INFO") ;

#ifndef SILENT
  cout << "TargetAnalyser (" << this << "): powering up..." << endl;
#endif
#ifdef RESOURCE_USAGE
  core = sbrk(0);
#endif
  states = 0;
  trans = 0;

  timer = new Timer;
  timer->start();
  info = new Information;
  if (!info->input(optionsFileName))
     terminate("cannot read options file");
  if (!info->inputStateInfo(infoFileName))
     terminate("cannot read state information file");
  tangible = info->getTangible();
  genRepInterval = info->getGenRepInterval();
  genRepStyle = info->getGenRepStyle();
  #ifndef SILENT
  cout << "TargetAnalyser (" << this << "): powered up..." << endl;
  #endif
}

void TargetAnalyser::destroy()
{
  #ifdef VERBOSE
  cout << "TargetAnalyser (" << this << "): destructor..." << endl;
  #endif
  delete timer;
  delete info;
  #ifndef SILENT
  cout << "TargetAnalyser (" << this << "): powered down..." << endl;
  #endif
}

TargetAnalyser::~TargetAnalyser()
{
  destroy();
}

void TargetAnalyser::search(char* basename)
{
  //long n;

  char stateBuffer[BUFFER_SIZE];
  char *filename = info->getStateListFile();
  FILE *sf = fopen(filename,"rb");
  if (!sf) terminate ("cannot open state file '%s'", filename);
  setvbuf(sf, stateBuffer, _IOFBF, BUFFER_SIZE);

  FullState current;
  current.initialise(info->getTransitions(), info->getTransType(), info->getPriority());

  long state = 0, targets = 0, initials = 0;

  Vector<int> target;
  Vector<double> weight, steady;

#ifdef USE_STEADY
  //  steady.input("STEADY");
  steady.input(info->getSteadyStateFile());
#endif

  target.setLength(0);

  while(current.input(sf)) {
#ifdef DEBUG
    cout << "inputted state: " << endl;
    current.output();
#endif

    weight.check(state);

#ifdef USE_STEADY
    if (current.passageWeight() > 0) {
      // cout << "using steady state " << steady[state] << endl;
      weight[state] = steady[state];
      initials++;
    } else
      weight[state] = 0.0;
#else
    weight[state] = current.passageWeight();
#endif
 
    if (current.isTarget()) {
      //      cout << "target state found (" << state << "): " << endl;
      //      current.output();
      target.check(targets);
      target[targets] = state;
      targets++;
    }
#ifdef DEBUG
#endif
    state++;
#ifndef SILENT
    if (genRepInterval && !(state % genRepInterval) && genRepStyle != NONE) 
      cout << "Reading state " << state << "..." << endl;
#endif
  }
  assert(state == tangible);

  fclose(sf);

  char* targetsFileName = add_extension(basename, "TARGETS") ;
  char* weightFileName  =  add_extension(basename, "WEIGHT") ;

  // output vector of targets to disk
  target.setLength(targets);
  target.output(targetsFileName);
#ifndef SILENT
  cout << "--> " << initials << " initial states found <--" << endl;
  cout << "--> " << targets << " target states found <--" << endl;
#endif
  weight.setLength(tangible);
#ifndef SILENT
  cout << "--> initial state weight sum is " << weight.oneNorm() << " <--" << endl;
#endif
  weight.normalise();
  weight.output(weightFileName);
}

// ************************

int main(int argc, char *argv[])
{
   if (argc != 2)
   {
      printf("Must supply the basename as an argument\n") ;
      printf("Number of arguments was: %d\n", argc) ;
      exit(1) ;
   }
  TargetAnalyser *target = new TargetAnalyser(argv[1]);

  target->search(argv[1]);

  delete target;
}


