
//  Program to launch the various parts of HYDRA-S in sequence
//  by Nick Dingle <njd200@doc.ic.ac.uk> 08/10/02
//  after "driver.cxx" by Will Knottenbelt

using namespace std;

#include <iostream>
#include <iomanip>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "common.hpp"


int fail() {
  return soap_sender_fault(soap, "File not found", NULL);
}

int execute(char *args[], int parameters) {
  for (int n=0; n<parameters; n++)
    cout << args[n] << " ";
  cout << endl;

  args[parameters] = NULL;
  long pid = fork();
  int status;
  if (pid==0) {
    execvp(args[0], args);
    cout << "execution failure (pid == 0)." << endl;
    exit(-1);
  }
  if (pid < 0) {
    cout << "execution failure (pid < 0)."  << endl;
    exit(-1);
  }
  while(1) {
    if (waitpid(pid, &status, 0)) {
      cout << "return code is " << status << endl;
      return !status;
    }
  }
}


int ns__InvokeHydra(struct soap *soap, char *filename, bool *success)
{
  // mimic main's argc and argv[]
  int argc = 2;
  char *argv[2];
  argv[0] = NULL;
  argv[1] = filename;

  // copy of the main() code from hydra-s-start here

  char *args[20];
  char *install = getenv("HYDRA_S_DIR");
  char libraries[512], headers[512];
  char usrdir[MAX_LENGTH]; 
  char libdir[MAX_LENGTH], hppdir[MAX_LENGTH];

  if (!install) {
    cout << argv[0] << ": set environment variable HYDRA_S_DIR to HYDRA-S's install directory" << endl;
    return 1;
  }

  if (argc < 2) {
    cout << argv[0] << ": no input file specified" << endl;
    return 1;
  }

  sprintf(libdir, "%s/lib", install);
  sprintf(hppdir, "%s/hpp", install);

 
  // parse <name>
  // make -f gspn-statemake
  // gen
  // make -f gspn-steadymake
  // gspn-steady
  // make -f gspn-targetmake
  // gspn-target
  // make -f gspn-passagemake


  cout << endl << "Parsing input file..." << endl;
  sprintf(usrdir, "%s/parse", install);
  //  args[0] = "parse";
  args[0] = usrdir;
  args[1] = argv[1];
  if (!execute(args, 2)) return fail();

  cout << endl;

  cout << endl << "Compiling user code..." << endl;
  args[0] = "g++";
  args[1] = "-O";
//  args[2] = "-O3";
  args[2] = "-c";
  sprintf(headers,"-I%s",hppdir);
  args[3] = headers;
  args[4] = "user.cxx";
  if (!execute(args, 5)) return fail();

  cout << endl << "Linking state generator..." << endl;
  args[0] = "g++";
  args[1] = "-O";
  args[2] = "user.o";
  sprintf(libraries,"-L%s",libdir);
  args[3] = libraries;
  args[4] = "-o";
  //  sprintf(usrdir, "%s/gen", install);
  //  args[5] = usrdir;
  args[5] = "gen";
  args[6] = "-lstate";
  args[7] = "-lcommon";
  args[8] = "-lm";
  args[9] = "-lstdc++";
  if(!execute(args, 10)) return fail();

  cout << endl << "Generating state space..." << endl;
  sprintf(usrdir, "%s/gen", install);
  args[0] = usrdir;
  if (!execute(args, 1)) return fail();

  cout << endl << "Performing functional analysis..." << endl;
  sprintf(usrdir, "%s/func", install);
  args[0] = usrdir;
  if (!execute(args, 1)) return fail();

  cout << endl << "Calling steady state analyser..." << endl;
  sprintf(usrdir, "%s/hydra-steady", install);
  args[0] = usrdir;
  if (!execute(args, 1)) return fail();

  cout << endl << "Linking target state identifier..." << endl;
  args[0] = "g++";
  args[1] = "-O";
  args[2] = "user.o";
  sprintf(libraries,"-L%s",libdir);
  args[3] = libraries;
  args[4] = "-o";
  //  sprintf(usrdir, "%s/gspn-target", install);
  //  args[5] = usrdir;
  args[5] = "gspn-target";
  args[6] = "-ltarget";
  args[7] = "-lcommon";
  args[8] = "-lm";
  args[9] = "-lstdc++";
  if(!execute(args, 10)) return fail();

  cout << endl;

  cout << "Identifying target states..." << endl;
  sprintf(usrdir, "%s/gspn-target", install);
  args[0] = usrdir;
  if (!execute(args, 1)) return fail();

  cout << endl;

  cout << "Linking performance analyser..." << endl;
  args[0] = "g++";
  args[1] = "-O";
  args[2] = "user.o";
  sprintf(libraries,"-L%s",libdir);
  args[3] = libraries;
  args[4] = "-o";
  //  sprintf(usrdir, "%s/perf", install);
  //  args[5] = usrdir;
  args[5] = "perf";
  args[6] = "-lperf";
  args[7] = "-lcommon";
  args[8] = "-lm";
  args[9] = "-lstdc++";
  if(!execute(args, 10)) return fail();

  cout << endl;

  sprintf(usrdir, "%s/perf", install);

  //  cout << "For steady-state analysis, execute \"perf\"" << endl;
  cout << "For steady-state analysis, execute \"" << usrdir << "\"" << endl;
  cout << "  -- results aren't written to disk, so use \"tee\" or pipe to a file to save them" << endl;

  cout << endl;

  sprintf(usrdir, "%s/uniform", install);

  //  cout << "For response-time/transient analysis, execute \"uniform\"" << endl;
  cout << "For response-time/transient analysis, execute \"" << usrdir << "\"" << endl;
  cout << "  -- use the \"-cdf\" option to produce the distribution rather than the density" << endl;
  cout << "  -- in either case there is no need to specify the model filename" << endl;
  cout << "  -- results are written to the file \"PT_RESULTS\" as <t-point> <value>" << endl;
  cout << "  -- or you can grep the output for \"DATA0\" to see the results as well" << endl;

  /*
  cout << "Compiling response-time analyser..." << endl;
  args[0] = "make";
  args[1] = "-f";
  sprintf(usrdir, "%s/uniformmake", install);
  args[2] = usrdir;
  if (!execute(args, 3)) return fail();
  */
  /*
  sprintf(usrdir, "%s/uniform", install);
  args[0] = usrdir;
  if (!execute(args, 1)) return fail();
  */

  cout << endl;
 
  cout << "Execution complete." << endl << endl; 

  // SOAP finalisation 
  *success = true;
  return SOAP_OK;
 
}


int main(int argc, char *argv[])
{ struct soap soap;
  int m, s;
  soap_init(&soap);
  // soap.accept_timeout = 60; // die if no requests are made within 1 minute
  if (argc < 2)
  { soap_serve(&soap);
    soap_destroy(&soap);
    soap_end(&soap);    // clean up
  }
  else
  { m = soap_bind(&soap, NULL, atoi(argv[1]), 100);
    if (m < 0)
    { soap_print_fault(&soap, stderr);
      exit(1);
    }
    fprintf(stderr, "Socket connection successful %d\n", m);
    for (int i = 1; ; i++)
    { s = soap_accept(&soap);
      if (s < 0)
      { if (soap.errnum)
          soap_print_fault(&soap, stderr);
        else
          fprintf(stderr, "Server timed out\n");
        break;
      }
      fprintf(stderr, "%d: accepted %d IP=%d.%d.%d.%d ... ", i, s, (int)(soap.ip>>24)&0xFF, (int)(soap.ip>>16)&0xFF, (int)(soap.ip>>8)&0xFF, (int)soap.ip&0xFF);
      soap_serve(&soap);        // process RPC skeletons
      fprintf(stderr, "served\n");
      soap_destroy(&soap);
      soap_end(&soap);  // clean up
    }
  }
  soap_done(&soap);
  return 0;
}
