////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

#define BUFFER_SIZE 131072
//#define VERBOSE

using namespace std;

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstdarg>
#include <cmath>
#include <complex>
#include <sys/types.h>
#include <ctime>
#include "common.hpp"
#include "dyna.hpp"
#include "timer.hpp"
#include "info.hpp"
#include "element.hpp"
#include "vector.hpp"
#include "store.hpp"
#include "avl.hpp"
#include "sparse-nocomplex.hpp"
#include "hydra-steady.hpp"
#include "filenames.hpp"
#ifdef RESOURCE_USAGE
#include <unistd.h>
#endif

void SteadyStateAnalyser::setRandomSeed() {
  long current_time;
  // if the line below gives compilation trouble, comment it out
  time(&current_time);
#ifdef SRAND48
  srand48(current_time);
#else
  srand(current_time);
#endif
}


#ifndef LOGGING
#define SILENT 1
#endif

int SteadyStateAnalyser::checkNegative(Vector<double> &solution)
{
  if (tangible < 2) return 0;
  long n, negative = 0;
  for (n=0; n<tangible; n++)
    if (solution[n] <= -EPS) negative++;
  if (negative > tangible - negative) {
    for (n=0; n<tangible; n++)
      solution[n] = -solution[n];
    return 1;
  }
  return 0;
}

int SteadyStateAnalyser::readColumnInfo() {
  char *filename = info->getTransMatrixFile();
  long n, states = 0; 
  int t; 

  #ifndef SILENT
  cout << "SteadyStateAnalyser (" << this << "): inspecting tangible states..." << endl;
  #endif

  colLength.setLength(tangible);

  for (n=0; n<tangible; n++)
    colLength[n] = 1;

  char buffer[BUFFER_SIZE];
  FILE *mf = fopen(filename,"rb");
  if (!mf) terminate("cannot open state matrix file '%s'", filename);
  setvbuf(mf, buffer, _IOFBF, BUFFER_SIZE);

  int kids;
  StateMatrixElement *child = new StateMatrixElement[info->getMaxChildren()];

  while(fread(&kids, sizeof(int), 1, mf)) {
    fread(child, sizeof(StateMatrixElement), kids, mf); 
    for (t=0; t<kids; t++)  {
      if (child[t].getDestination() != states)
        colLength[child[t].getDestination()]++;
    }
    states++;
    #ifndef SILENT
    if (genRepInterval && !(states % genRepInterval) && genRepStyle != NONE) cout << "Reading state " << states << "..." << endl;
    #endif
  }
  assert(states == tangible);

  for (n=0; n<tangible; n++)
    Q->getCol(n).setLength(colLength(n));

  colLength.setLength(0);
  delete [] child;
  fclose(mf);
  #ifndef SILENT
  cout << "SteadyStateAnalyser (" << this << "): inspect complete ( " << states << " states)" << endl;
  #endif
  return 1;
}

int SteadyStateAnalyser::readStates()
{
  #ifndef SILENT
  cout << "SteadyStateAnalyser (" << this << "): reading tangible states...\n";
  #endif
  char *filename = info->getTransMatrixFile();

  timer->start(); 
  scale.setLength(tangible);
  switch(solver) {
  case ROW_COLUMN:
    readStatesQRC(filename);
    break;
  case COLUMN:
    readStatesQC(filename);
    break;
  case DIRECT:
    readStatesQD(filename);
    break;
  default:
    terminate("internal error (bad solver type %d)", solver);
    break;
  }
  #ifndef SILENT
  cout << "SteadyStateAnalyser (" << this << "): read complete ( " << tangible << " states, " << trans << " transitions)" << endl;
  #endif
  #ifndef SILENT
  double timeUsed = timer->read();
  if (timeUsed) {
    cout << "(read speed " << ((double) tangible/timeUsed) << " states/second, " << ((double) trans/timeUsed) << " transitions/second)\n";
  }
  if (solRepStyle != NONE) Q->statistics();
  cout << endl;
  #endif
  return 1;
}

int SteadyStateAnalyser::readStatesQRC(char *filename) {
  long dest, states = 0;
  int t; 
  double weight, sum, factor;

  char buffer[BUFFER_SIZE];
  FILE *mf = fopen(filename,"rb");
  if (!mf) terminate("cannot open state matrix file '%s'", filename);
  setvbuf(mf, buffer, _IOFBF, BUFFER_SIZE);

  int kids;
  StateMatrixElement *child = new StateMatrixElement[info->getMaxChildren()];

  while(fread(&kids, sizeof(int), 1, mf)) {
    fread(child, sizeof(StateMatrixElement), kids, mf); 
    
    sum = 0;
    for (t=0; t<kids; t++)  {
      if (child[t].getDestination() != states)
	sum += child[t].getWeight();
    }
    
    factor = (double) 1.0/sum;
    for (t=0; t<kids; t++) {
      dest = child[t].getDestination();
      weight = child[t].getWeight()*factor;
      if (dest != states)
	QRC->addElement(states, dest, -weight);
    }
    
    QRC->addElement(states, states, 1.0);

    if(!doSteady) {
      scale[states] = 1.0; // factor;
    } else {
      scale[states] = factor;
    }

    QRC->getRow(states).lock();
    trans += kids;
    states++;
#ifndef SILENT
    if (genRepInterval && !(states % genRepInterval) && genRepStyle !=NONE) cout << "Reading state " << states << "..." << endl;
#endif
  }
  assert(states == tangible);
  delete [] child;
  fclose(mf); 
    
  return 1;
}

int SteadyStateAnalyser::readStatesQC(char *filename) {
  long dest, n, states = 0;
  int t; 
  double weight, sum, factor;

  char buffer[BUFFER_SIZE];
  FILE *mf = fopen(filename,"rb");
  if (!mf) terminate("cannot open state matrix file '%s'", filename);
  setvbuf(mf, buffer, _IOFBF, BUFFER_SIZE);

  int kids;
  StateMatrixElement *child = new StateMatrixElement[info->getMaxChildren()];
  assert(info->getMethod() != AUTOMATIC);
  if (info->getMethod() != AI && info->getMethod() != AIR && info->getMethod() != POWER) {
    while(fread(&kids, sizeof(int), 1, mf)) {
      fread(child, sizeof(StateMatrixElement), kids, mf); 
      
      sum = 0;
      for (t=0; t<kids; t++)  {
	if (child[t].getDestination() != states)
	  sum += child[t].getWeight();
      }
      
      factor = (double) 1.0/sum;
      for (t=0; t<kids; t++) {
	dest = child[t].getDestination();
	weight = child[t].getWeight()*factor;
	if (dest != states) {
//          cout << "adding (" << states << "," << dest << " = " << weight << ")" << endl;
	  QC->addElement(states, dest, weight);
//          assert(QC->getElement(states,dest)==weight);
	}
      }
      QC->addElement(states, states, -1.0);
//      assert(QC->getElement(states, states)==-1.0);
//      assert(QC->getElement(0,0)==-1.0);

      if(!doSteady) {
	scale[states] = 1.0; // factor;
      } else {
	scale[states] = factor;
      }

      trans += kids;
      states++;
#ifndef SILENT
      if (genRepInterval && !(states % genRepInterval) && genRepStyle !=NONE) cout << "Reading state " << states << "..." << endl;
#endif
    }
  } else if (info->getMethod() != POWER) {
    ai_b.setLength(tangible);
    ai_d.setLength(tangible);
    ai_start = 0.0;

    for (n = 0; n<tangible; n++) {
      ai_d[n] = 0.0;
      ai_b[n] = 0.0;
    }

    while(fread(&kids, sizeof(int), 1, mf)) {
      fread(child, sizeof(StateMatrixElement), kids, mf); 
      
      sum = 0;
      for (t=0; t<kids; t++) 
	  sum += child[t].getWeight();
      
      factor = (double) 1.0/sum;
      for (t=0; t<kids; t++) {
	dest = child[t].getDestination();
	weight = child[t].getWeight()*factor;
        if (dest > states)
	  ai_b[states] += weight;
	else if (dest < states)
	  ai_d[states] += weight;
        QC->addElement(states, dest, weight);
      }

      if (!states) {
        for (t=0; t<kids; t++) {
  	  dest = child[t].getDestination();
	  weight = child[t].getWeight()*factor;
          if (dest > 1) ai_start += weight;
	}
      }

      if(!doSteady) {
	scale[states] = 1.0; // factor;
      } else {
	scale[states] = factor;
      }

      trans += kids;
      states++;
#ifndef SILENT
      if (genRepInterval && !(states % genRepInterval) && genRepStyle !=NONE) cout << "Reading state " << states << "..." << endl;
#endif
    }
  } else { // POWER_METHOD
    double ALPHA;
    cout << "ALPHA: " << endl;
    cin >> ALPHA;
    int diag;
    cout << "!!! READING POWER !!!" << endl;
    while(fread(&kids, sizeof(int), 1, mf)) {
      fread(child, sizeof(StateMatrixElement), kids, mf); 
      
      sum = 0;
      for (t=0; t<kids; t++)
	sum += child[t].getWeight();
      
      factor = (double) 1.0/sum;
      diag = 1;
      for (t=0; t<kids; t++) {
	dest = child[t].getDestination();
	weight = child[t].getWeight()*factor;
	if (dest == states) {
	  diag = 0;
    	  QC->addElement(states, dest, 1.0-ALPHA+weight*ALPHA);	  
	} else {
	  QC->addElement(states, dest, weight*ALPHA);
	}
      }      
      if (diag) QC->addElement(states, states, 1.0-ALPHA);

      if(!doSteady) {
	scale[states] = 1.0; // factor
      } else {
	scale[states] = factor;
      } 

      trans += kids;
      states++;
#ifndef SILENT
      if (genRepInterval && !(states % genRepInterval) && genRepStyle !=NONE) cout << "Reading state " << states << "..." << endl;
#endif
    }
  }
  assert(states == tangible);
  delete [] child;
  fclose(mf);
  return 1;
}

int SteadyStateAnalyser::readStatesQD(char *filename) {
  long dest, states;
  int t; 
  double weight, sum;

  states = 0;

  char buffer[BUFFER_SIZE];
  FILE *mf = fopen(filename,"rb");
  if (!mf) terminate("cannot open state matrix file '%s'", filename);
  setvbuf(mf, buffer, _IOFBF, BUFFER_SIZE);

  int kids;
  StateMatrixElement *child = new StateMatrixElement[info->getMaxChildren()];

  while(fread(&kids, sizeof(int), 1, mf)) {
    fread(child, sizeof(StateMatrixElement), kids, mf); 

    sum = 0;
    for (t=0; t<kids; t++) {
      dest = child[t].getDestination();
      weight = child[t].getWeight();
      if (dest != states) {
        sum += weight;
        QC->addElement(states, dest, weight);
      }
    }
    scale[states] = 1.0/sum;
    QC->addElement(states, states, -sum);
//    QC->getRow(states).lock();   
    trans += kids;
    states++;
    #ifndef SILENT
    if (genRepInterval && !(states % genRepInterval) && genRepStyle != NONE) cout << "Reading state " << states << "..." << endl;
    #endif
  }
  assert(states == tangible);
  delete [] child;
  fclose(mf);
  return 1;
}

void SteadyStateAnalyser::unScale(Vector<double> &solution)
{
  for (long n=0; n<tangible; n++)
    solution[n] *= scale(n);
}

void SteadyStateAnalyser::reScale(Vector<double> &solution)
{
  for (long n=0; n<tangible; n++)
    solution[n] /= scale(n);
}

SteadyStateAnalyser::SteadyStateAnalyser(char* optionsFileName, char* infoFileName)
{

#ifndef SILENT
  cout << "SteadyStateAnalyser (" << this << "): powering up...\n";
#endif
  Q = NULL;
#ifdef RESOURCE_USAGE
  core = sbrk(0);
#endif
  trans = 0;

  doSteady = doPassage = doTransient = 0;

  info = new Information;
  timer = new Timer;
  if (!info->input(optionsFileName)) 
     terminate("hydra-steady.cxx cannot read options file");
  if (!info->inputStateInfo(infoFileName)) 
     terminate("hydra-steady.cxx cannot read state information file");
  genRepInterval=  info->getGenRepInterval();
  genRepStyle = info->getGenRepStyle();
  solRepInterval = info->getSolRepInterval();
  solRepStyle = info->getSolRepStyle();
  maxIterations = info->getMaxIterations();
  accuracy = info->getAccuracy();
  relaxParameter = info->getRelaxParameter();
  tangible = info->getTangible();
  steadyVectorFile = info->getSteadyStateFile();
  if (!info->getGeneration()) terminate ("run state space generator first");
  if (!info->getTangible()) terminate("no tangible states");
  if (!info->getFunctional())
    cout << "warning: no functional analysis has been performed on model. recommend" << endl << "         running functional analyser before proceeding" << endl << endl;
  if (info->getDeadlock())
    terminate ("steady state solution is not possible since model contains deadlock");
  if (info->getReducible())
    terminate ("steady state solution is not possible since model is reducible");
  solver = UNKNOWN;
  QD = NULL;
  QRC = NULL;
  QC = NULL;

  doSteady = info->getDoSteady();
  doPassage = info->getPassage();
  doTransient = info->getTransient();

  if(doSteady)
    assert(!doPassage && !doTransient);
 

  SolutionMethod method = info->getMethod();

  if (method==AUTOMATIC) {
    if (tangible <= 250) 
      method = GRASSMANN;
    else if (tangible <= 500)
      method = GAUSS;
    else if (tangible <= 20000) {
      method = SOR;
      info->setDynamic(1);
    } else if (tangible <= 50000) {
      method = CGS;
    } else
      method = AIR;
  }
  info->setMethod(method);

  assert(info->getMethod()!=AUTOMATIC);

  switch(info->getMethod()) {
  case GAUSS:
  case LU:
  case GRASSMANN:
    solver = DIRECT;
    QC = new SparseMatrix(tangible, tangible);
    Q = QC;
    break;    
  case CGNR:
  case BICG:
    solver = ROW_COLUMN;
    QRC = new SparseMatrixWithRowColumn(tangible, tangible);
    Q = QRC;
    break;
  case POWER:
  case GAUSS_SEIDEL:
  case SOR:
  case CGS:
  case BICGSTAB:
  case BICGSTAB2:
  case MYCG:
  case AI:
  case AIR:
    solver = COLUMN;
    QC = new SparseMatrix(tangible, tangible);
    Q = QC;
    break;
    /*  case AUTOMATIC:
    if (tangible <= 500) {
      solver = DIRECT;
      QC = new SparseMatrix(tangible, tangible);
      Q = QC;
    } else {
      solver = COLUMN;
      QC = new SparseMatrix(tangible, tangible);
      Q = QC;
    }
    break;*/
  default:
    terminate("internal error (bad solution method %d)", info->getMethod());
    break;
  }

  setRandomSeed();
  #ifndef SILENT
  cout << "SteadyStateAnalyser (" << this << "): powered up...\n";
  #endif
}

void SteadyStateAnalyser::destroyQ()
{
  if (Q) delete Q;
}

void SteadyStateAnalyser::destroy()
{ 
  #ifdef VERBOSE
  cout << "SteadyStateAnalyser (" << this << "): destructor...\n";
  #endif
  delete info;
  delete timer;
  cout.unsetf(ios::showpoint);
  #ifndef SILENT
  cout << "SteadyStateAnalyser (" << this << "): powered down...\n";
  #endif
}

SteadyStateAnalyser::~SteadyStateAnalyser()
{
  destroy();
}

void SteadyStateAnalyser::terminate(char *fmt, ...)
{
  char buffer[2048];

  va_list argp;
  va_start(argp, fmt);
  vsprintf(buffer, fmt, argp); 
  va_end(argp);
  destroy();
  printf("\nfatal error: %s.\nexecution halted.\n", buffer);  
  exit(1);
}

void SteadyStateAnalyser::solve(char* infoFileName)
{
  /* common parameters: max iterations, accuracy */
  Vector<double> *soln;

  int startVector = strlen(info->getStartVectorFile());

  if (startVector) {
    soln = new Vector<double>;
    if (!soln->input(info->getStartVectorFile()))
      terminate ("error reading steady state file '%s'", info->getStartVectorFile());
    if (soln->getLength() != tangible)
      terminate ("given steady state vector has %d elements (should have %d)", soln->getLength(), tangible);
    for (long n=0; n<tangible; n++)
      (*soln)[n] /= scale(n);
    soln->normalise();
  } else
    soln = new Vector<double>(tangible);

  Vector<double> &solution = *soln;

  if (!startVector) {
    long n;
    switch(info->getMethod()) {
    case GAUSS:
    case LU:
    case GRASSMANN:
      break;
    case GAUSS_SEIDEL:
    case SOR:
      for (n=0; n<tangible; n++)
	solution[n] = 1;
      break;
    case BICG:
    case CGS:
    case BICGSTAB:
    case BICGSTAB2:
    case CGNR:
    case MYCG:
    case POWER:
#ifdef SRAND48
      for (n = 0; n < tangible; n++)
	solution[n] = 25.0*drand48();
#else
      for (n = 0; n < tangible; n++)
	solution[n] = 25.0*rand()/32767.0;
#endif
      break;
    case AI:
    case AIR:
      solution[0] = 1.0/tangible;
      for (n = 1; n < tangible; n++)
	solution[n] = solution(0);
      break;
    case AUTOMATIC:
      terminate("internal error (AUTOMATIC not handled)");
      break;
    }
  }

  switch(info->getMethod()) {
  case GAUSS:
    gauss(solution);
    break;
  case LU:
    break;
    case GRASSMANN:
    grassmann(solution);
    break;
  case POWER:
    power(solution, maxIterations, accuracy);
    break;
  case GAUSS_SEIDEL:
    gaussSeidel(solution, maxIterations, accuracy);
    break;
  case SOR:
  if (info->getDynamic())
    dynamicSOR(solution, maxIterations, accuracy);
  else
    sor(solution, maxIterations, accuracy);
    break;
  case CGNR:
    cgnr(solution, maxIterations, accuracy);
    break;
  case CGS:
    cgs(solution, maxIterations, accuracy);
    break;
  case BICG:
    bicg(solution, maxIterations, accuracy);
    break;
  case MYCG:
    tfqmr(solution, maxIterations, accuracy);
    break;
  case BICGSTAB:
    bicgstab(solution, maxIterations, accuracy);
    break;
  case BICGSTAB2:
    bicgstab2(solution, maxIterations, accuracy);
    break;
  case AI:
    ai(solution, maxIterations, accuracy, NO_REFINE);
    break;
  case AIR:
    air(solution, maxIterations, accuracy, NO_REFINE);
    break;
  default:
    terminate("internal error (bad method type %d)", info->getMethod());
  }
#ifdef RESOURCE_USAGE
  unsigned long memory = (unsigned long) (sbrk(0) - core);
  cout << "SteadyStateAnalyser (" << this << "): " << (double) (memory/1024.0) << " KB memory used for data." << endl;
#endif  
  info->setSteady(1);
  if (!info->outputStateInfo(infoFileName))
    terminate("cannot update state information file");
  delete soln;
}

long SteadyStateAnalyser::checkVector(Vector<double> &solution, double infNorm)
{

  long n;

  for (n=0; n<tangible; n++)
    if (solution(n) < 0 && solution(n) > -EPS) solution(n) = 0;

  double oneNorm = solution.oneNorm();
  long status = 0;
  double min, max;
  cout.setf(ios::showpoint);
  cout.setf(ios::scientific);

#ifdef LOGGING
  cout << "final steady state vector characteristics: " << endl << endl;
  cout << "  normalised residual norm.........." << setw(19) << infNorm << endl;
  cout << "  vector elements sum to............" << setw(19) << oneNorm << endl;
#endif

  min = solution(0);
  max = solution(0);
  register double current;
  for (n=1; n<tangible; n++) {
    current = solution(n);
    if (current < min) min = current;
    if (current > max) max = current;
    if (!(current > 0) && !(current<=0)) status |= MY_NAN;
  }

#ifdef LOGGING
  int prec = cout.precision(12);
  cout << "  all vector elements are in range.." << setw(19) << min << " to " << setw(19) << max << endl;
  cout.unsetf(ios::showpoint);
  cout.unsetf(ios::scientific);
  cout.precision(prec);
#endif

  if (infNorm > 10000*accuracy) {
// ###
//    cout << "large residual detected!!!" << endl;
    status = (LARGE_NORM | status);
//    cout << "LARGE NORM = " << LARGE_NORM << endl;
//    cout << "LARGE_NORM | status = " << (LARGE_NORM | status) << endl;
//    cout << "status is " << status << endl;
  }
  if (min < 0 || max > 1.0) status = (RANGE_VIOLATION | status);
  if (fabs(oneNorm-1.0)>1e-10) status = (SUM_VIOLATION | status);

#ifdef LOGGING
  cout << "  steady state vector status........";
  if (status)
    cout << " ***FAILED***";
  else
    cout << " pass";
  cout << endl << endl;
// ###
  cout << "returning status: " << status << endl;
#endif
  return status;
}

void SteadyStateAnalyser::checkViolation(long status)
{
// ###
  cout << "status is " << status << endl;

  if (status & MY_NAN) terminate("not-a-number detected in steady state vector");
  if (status & LARGE_NORM)
    terminate("unacceptably large normalised residual norm");
  if (status & SUM_VIOLATION)
    terminate("steady state vector element sum violation");
  if (status & RANGE_VIOLATION)
    terminate("steady state vector element range violation");
}

long SteadyStateAnalyser::refine(Vector<double> &_solution, long _iterations, double _accuracy)
{
  cout << "refining the solution..." << endl;

  info->setMethod(AI);
  if (QC) delete QC;
  if (QD) delete QD;
  if (QRC) delete QRC;
  QC = new SparseMatrix(tangible, tangible);
  Q = QC;
  solver = COLUMN;
  readColumnInfo();
  readStates();
  ai(_solution, _iterations, _accuracy, REFINE);
  return 1;
}


int main(int argc, char *argv[])
{
   if (argc != 2)
   {
      printf("Must supply the basename as an argument\n") ;
      printf("Number of arguments was: %d\n", argc) ;
      exit(1) ;
   }

   char* basename        = argv[1] ;
   char* optionsFileName = add_extension(basename, "OPTIONS") ;
   char* infoFileName    =  add_extension(basename, "INFO") ;

  SteadyStateAnalyser *steady = new SteadyStateAnalyser(optionsFileName, infoFileName);
  steady->readColumnInfo();
  steady->readStates();
  steady->solve(infoFileName);
  steady->destroyQ();
  delete steady;
}

