////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

//#define DEBUG
//#define BUFFER_SIZE 16384

using namespace std;

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstdarg>
#include <cmath>
//#include <complex.h>
#include "common.hpp"
#include "dyna.hpp"
#include "timer.hpp"
#include "info.hpp"
#include "element.hpp"
#include "vector.hpp"
#include "user.hpp"
#include "full.hpp"
#include "perform.hpp"
#include "filenames.hpp"
#ifdef RESOURCE_USAGE
#include <unistd.h>
#endif

void ValueList::insert(double _dbnValue, double _dbnFrequency)
{
  ValueListEntry *item = new ValueListEntry(_dbnValue, _dbnFrequency);
  if (!front) {
    front = item;
    return;
  }
  ValueListEntry *p, *q;
  p = front;
  q = NULL;
  while (p && _dbnValue > p->getDbnValue()) {
    q = p;
    p = p->getNext();
  }
  if (!q) { // insert at front
    item->setNext(front);
    front = item;
    return;
  }
  item->setNext(p);
  q->setNext(item);
}

int StateMeasure::showDistribution() {
  ValueList *valueList = new ValueList;
  for (int t=0; t<values; t++)
    valueList->insert(dbnValues(t),dbnFrequency(t));
  valueList->output();
  delete valueList;
  return 1;
}

void PerformanceAnalyser::terminate(char *fmt, ...)
{
  char buffer[2048];

  va_list argp;
  va_start(argp, fmt);
  vsprintf(buffer, fmt, argp); 
  va_end(argp);
  destroy();
  printf("\nfatal error: %s.\nexecution halted.\n", buffer);  
  exit(1);
}

int PerformanceAnalyser::readFinalSolution() {
  char *filename = info->getSteadyStateFile();
  final.input(filename);
  //  final.input("SM-STEADY");
#ifdef DEBUG
  for (long n=0; n<tangible; n++) 
    cout << "final [" << n << "] = " << final(n) << endl;
#endif
  return 1;
}

PerformanceAnalyser::PerformanceAnalyser(char* basename)
{
   char* optionsFileName = add_extension(basename, "OPTIONS") ;
   char* infoFileName    = add_extension(basename, "INFO") ;
#ifndef SILENT
  cout << "PerformanceAnalyser (" << this << "): powering up..." << endl;
#endif
#ifdef RESOURCE_USAGE
  core = sbrk(0);
#endif
  states = 0;
  trans = 0;

  timer = new Timer;
  timer->start();
  info = new Information;
  if (!info->input(optionsFileName))
     terminate("perform.cxx cannot read options file");
  if (!info->inputStateInfo(infoFileName)) 
     terminate("perform.cxx cannot read state information file");
  tangible = info->getTangible();
  genRepInterval = info->getGenRepInterval();
  genRepStyle = info->getGenRepStyle();
  if (!info->getSteady()) terminate ("run steady state solver first");
  #ifndef SILENT
  cout << "PerformanceAnalyser (" << this << "): powered up..." << endl;
  #endif
}

void PerformanceAnalyser::destroy()
{
  #ifdef VERBOSE
  cout << "PerformanceAnalyser (" << this << "): destructor..." << endl;
  #endif
  delete timer;
  delete info;
  #ifndef SILENT
  cout << "PerformanceAnalyser (" << this << "): powered down..." << endl;
  #endif
}

PerformanceAnalyser::~PerformanceAnalyser()
{
  destroy();
}

void PerformanceAnalyser::computeMeasures()
{
  long n;

  char stateBuffer[BUFFER_SIZE];
  char *filename = info->getStateListFile();
  FILE *sf = fopen(filename,"rb");
  if (!sf) terminate ("cannot open state file '%s'", filename);
  setvbuf(sf, stateBuffer, _IOFBF, BUFFER_SIZE);

  FullState current;
  current.initialise(info->getTransitions(), info->getTransType(), info->getPriority());

  int stateMeasures = info->getStateMeasures();
  int countMeasures = info->getCountMeasures();

  int max;
  StateMeasure *stateMeasure = new StateMeasure[stateMeasures];
  double *countMeasure = new double[countMeasures];
  if (stateMeasures > countMeasures) max = stateMeasures;
    else max = countMeasures;
  double *measure = new double[max];

  for (n=0; n<countMeasures; n++)
    countMeasure[n] = 0;

  long state = 0;

  while(current.input(sf)) {
#ifdef DEBUG
    cout << "inputted state: " << endl;
    current.output();
#endif
    current.computeStateMeasures(measure);
#ifdef DEBUG
    cout << "final[" << state << "] is " << final[state] << endl;
#endif
    for (n=0; n<stateMeasures; n++)
      stateMeasure[n].add(measure[n], final[state]);

    for (n=0; n<countMeasures; n++)
      measure[n] = 0;

//    cout << "0!\n";
    current.findEnabledTransitions();

//    cout << "1!\n";
    current.calculateCountMeasures(measure);

//    cout << "2!\n";
    for (n=0; n<countMeasures; n++) {
      #ifdef DEBUG
      if (measure[n]) cout << "final state is: " << final[state] << endl;
      #endif
      countMeasure[n] += final[state] * measure[n];
    }
    state++;
#ifndef SILENT
    if (genRepInterval && !(state % genRepInterval) && genRepStyle != NONE) 
      cout << "Reading state " << state << "..." << endl;
#endif
  }
  assert(state == tangible);
  int acc = int(-log10(info->getAccuracy())+0.5);
  cout.precision(acc);
  cout << "(precision set to " << acc << " places)" << endl << endl;

  cout << "(begin performance results)" << endl << endl;
  int est;
  for (n=0; n<stateMeasures; n++) {
    cout << "State Measure '" << info->getStateName(n) << "'" << endl << endl;;
    est = info->getStateEstimator(n);
    cout.setf(ios::scientific);
    if (est & MEAN)
      cout << "  mean                 " << stateMeasure[n].getSum() << endl;
    if (est & VARIANCE)
      cout << "  variance             " << stateMeasure[n].getVariance() << endl;
    if (est & STDDEV)
      cout << "  std dev.             " << stateMeasure[n].getStdDev() << endl;
    if (est & DISTRIBUTION) {
      cout << "  distribution" << endl;
      stateMeasure[n].showDistribution();
    }
    cout << endl;
  }
  for (n=0; n<countMeasures; n++) {
    cout << "Count Measure '" << info->getCountName(n) << "'" << endl << endl;
    est = info->getCountEstimator(n);
    if (est & MEAN)
      cout << "  mean                 " << countMeasure[n] << endl;
    cout << endl;
  }
  cout.unsetf(ios::scientific);
  cout << "(end performance results)" << endl;
#ifdef RESOURCE_USAGE
  unsigned long memory = (unsigned long) (sbrk(0) - core);
  cout << "PerformanceAnalyser (" << this << "): " << (double) (memory/1024.0) << " KB memory used for data." << endl;
#endif
  fclose(sf);
  if (stateMeasures) delete [] stateMeasure;
  if (countMeasures) delete [] countMeasure;
  if (max) delete [] measure;
}

// ************************


int main(int argc, char *argv[])
{
   if (argc != 2)
   {
      printf("Must supply the options file as an argument\n") ;
      printf("Number of arguments was: %d\n", argc) ;
      exit(1) ;
   }

  PerformanceAnalyser *perform = new PerformanceAnalyser(argv[1]);

  perform->readFinalSolution();
  perform->computeMeasures();
  delete perform;
}


