class SparseMatrixRow {
protected:
  long entries;
  FastDynaArray<SteadyStateRowElement> element;
public:
  SparseMatrixRow(long n) { entries = n; element.setLength(n); }
  SparseMatrixRow() { entries = 0; }
  ~SparseMatrixRow() {} 
  long addElement(long column, double *index);
  long safeAddElement(long column, double *index);
  inline long getColumn(long index) const { return element(index).getColumn(); } 
  inline double *getIndex(long index) const { return element(index).getIndex(); }
  inline long getEntries() const { return entries; }
  inline void lock() { element.setLength(entries); }
  inline void setLength(long n) { element.setLength(n); }
  inline void remove() { element.destroy(); entries = 0;}
};

class SparseMatrixCol {
protected:
  long entries;
  FastDynaArray<SteadyStateColElement> element;
public:
  SparseMatrixCol(long n) { entries = n; element.setLength(n); }
  SparseMatrixCol() { entries = 0; }
  ~SparseMatrixCol() {} 
  void setLength(long n) { element.setLength(n); }
  inline void addElement(long row, double *index) {
    element.check(entries);
    element[entries].setRow(row);
    element[entries].setIndex(index);
    entries++;
  }
  inline long safeAddElement(long row, double *index) {
    for (long n=0; n<entries; n++) {
      if (element(n).getRow() == row) {
	element(n).setIndex(index);
	return 0;
      }
    }
    element.check(entries);
    element[entries].setRow(row);
    element[entries].setIndex(index);
    entries++;
    return 1;
  }

  inline long getRow(long index) const { return element(index).getRow(); }
  inline double *getIndex(long index) const { return element(index).getIndex(); }
  inline long getEntries() const { return entries; }
  inline void lock() { element.setLength(entries); }
  inline void remove() { element.destroy(); entries=0;}
};

class SparseMatrix {
protected:
  long rows, cols, elements, calls;
  FastDynaArray<SparseMatrixCol> col;
  AVLTree *tree;
  Store *store;
public:
  SparseMatrix(long r, long c);
  ~SparseMatrix();
  void statistics();
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getCalls() const { return calls; }
  inline long getElements() const { return elements; }
  inline SparseMatrixCol &getCol(long index) { return col(index); }

  inline double *insert(double _entry) {
    #ifdef DEBUG
    cout << "insert: searching for " << _entry << endl;
    #endif
    calls++;
    double *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
/*      // ###
      assert(_index == tree->search(_entry));*/
      elements++;
      //      cout << "store now has " << elements  << " elements" << endl;
    }
    return _index;
  }

  inline void addElement(long _row, long _column, double _entry) {
    double *_index = insert(_entry);    
    col[_column].addElement(_row, _index);
  }

  inline double getElement(long _row, long _column) const {
    for (long n=0; n<col(_column).getEntries(); n++)
      if (col(_column).getRow(n) == _row) return *(col(_column).getIndex(n));
    return 0.0;     
  }

  inline double getColElement(SparseMatrixCol *_col, long _index) const {
    return *(_col->getIndex(_index));
  }

  int invert();

  void transMultiply(Vector<double> &x, Vector <double> &result) const;
  void preMultiply(Vector<double> &x, Vector<double> &result) const;
  void calcResidual(Vector<double> &x, Vector <double> &result) const;
  void SORIteration(Vector<double> &x, double omega) const;
  void safeSORIteration(Vector<double> &x, double omega) const;
  void twinSORIteration(Vector<double> &x1, double omega1, Vector<double>
&x2, double omega2) const;
  void calcResidualNorm(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const;
  void calcResidualNormAI(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const;
  void twinCalcResidualNorm(Vector<double> &x1, double &x1Norm, double &x1InfNorm,
Vector<double> &x2, double &x2Norm, double &x2InfNorm, Vector<double> &last,
double &x1DiffNorm, double &x2DiffNorm) const;

  void print();
};

class SparseMatrixWithRow {
protected:
  long rows, cols, elements, calls;
  FastDynaArray<SparseMatrixRow> row;
  AVLTree *tree;
  Store *store;
public:
  SparseMatrixWithRow(long r, long c);
  ~SparseMatrixWithRow();
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getCalls() const { return calls; }
  inline long getElements() const { return elements; }
  inline SparseMatrixRow &getRow(long index) { return row(index); }

  inline double *insert(double _entry) {
    #ifdef DEBUG
    cout << "insert: searching for " << _entry << endl;
    #endif
    calls++;
    double *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
      elements++;
      //      cout << "store now has " << elements << " elements" << endl;
    }
    return _index;
  }

  inline void addElement(long _row, long _column, double _entry) {
    double *_index = insert(_entry);    
    row[_row].addElement(_column, _index);
  }

  inline int safeAddElement(long _row, long _column, double _entry) {
    double weight = getElement(_row, _column);
    double *_index = insert(_entry + weight);    
    return row[_row].safeAddElement(_column, _index);
  }

  inline double getElement(long _row, long _column) const {
    for (long n=0; n<row(_row).getEntries(); n++)
      if (row(_row).getColumn(n) == _column) return *(row(_row).getIndex(n));
    return 0.0;     
  }

  inline double getRowElement(SparseMatrixRow *_row, long _index) const {
    return *(_row->getIndex(_index));
  }

  void multiply(Vector<double> &x, Vector <double> &result) const;
  void inner(Vector<double> &x, Vector <double> &fixed) const;
  void innerSOR(Vector<double> &x, Vector <double> &fixed, double omega) const;
  void innerSOR(Vector<double> &x, Vector <double> &fixed, double &conv, double omega) const;
  void innerSOR(Vector<double> &x, Vector <double> &fixed, int lower, int upper, double omega) const;
  void innerSOR(Vector<double> &x, Vector <double> &fixed, double &maxSol, double &maxDiff, int lower, int upper, double omega) const;
  void integrity(char *message) const;
  void inner(Vector<double> &x, Vector <double> &fixed, double &conv) const;
  void print();
};


class SparseMatrixWithRowColumn : public SparseMatrix {
protected:
  FastDynaArray<SparseMatrixRow> row;
  int non_zeros;

public:
  SparseMatrixWithRowColumn(long r, long c) : SparseMatrix (r,c) {
//    cout << "SMWRC!" << endl;
    row.setLength(rows);
//    cout << "SMWRC constructed!" << endl;
    non_zeros = 0;
  }

  void removeState(int target) {

    cout << "target for removal=" << target << endl;
    //    row[target].setLength(0);
    //    col[target].setLength(0);

    row[target].remove();
    col[target].remove();

  }

  void multiply(Vector<double> &x, Vector<double> &result) const;
  inline void addElement(long _row, long _column, double _entry) {
    double *_index = insert(_entry);    
    row[_row].addElement(_column, _index);
    col[_column].addElement(_row, _index);    
    non_zeros++;
  }

  inline int safeAddElement(long _row, long _column, double _entry) {
    double weight = getElement(_row, _column);
    if (weight != 0)
      non_zeros++;
    double *_index = insert(_entry + weight);    
    int result = row[_row].safeAddElement(_column, _index);
    col[_column].safeAddElement(_row, _index);
    return result;
  }

  inline SparseMatrixRow &getRow(long index) { return row(index); }

  inline double getRowElement(long _row, long _index) const {
    return *(row(_row).getIndex(_index));
  }

  inline double getRowElement(SparseMatrixRow *_row, long _index) const {
//    double *index = _row->getIndex(_index);
    return *(_row->getIndex(_index));
  }

/*  inline void setElement(long _row, long _column, double _entry) {
    long _index = insert(_entry);
    row[_row].addElement(_column, _index);
    col[_column].addElement(_row, _index);
  }*/

  void print();
};

class DirectMatrix : public SparseMatrix { 
protected:
  FastDynaArray<SparseMatrixRow> row;
public:
  DirectMatrix(long r, long c) : SparseMatrix (r,c) {
    row.setLength(rows);
  }
  inline SparseMatrixRow &getRow(long index) { return row(index); }
  inline double *insert(double _entry) {
    calls++;
    double *_index = store->insert(_entry);
    elements++;
    return _index;
  }

  inline void set(double *_index, double _entry) {
    *_index = _entry;
//    store->setEntry(_index, _entry);
  }

  inline void addElement(long _row, long _column, double _entry) {
    double *_index = insert(_entry);
    row[_row].addElement(_column, _index);
    col[_column].addElement(_row, _index);
  }

  inline double getElement(long _row, long _column) const {
    for (long n=0; n<row(_row).getEntries(); n++)
      if (row(_row).getColumn(n) == _column) return *(row(_row).getIndex(n));
    return 0.0;     
  }

  inline double *getIndex(long _row, long _column) const
  {
    for (long n=0; n<row(_row).getEntries(); n++)
      if (row(_row).getColumn(n) == _column) return row(_row).getIndex(n);
    return NULL;
  }  

  inline void setElement(long _row, long _column, double _entry) {
    double *_index = getIndex(_row, _column);
    if (_index == NULL) {
      if (!_entry) return;
      _index = insert(_entry);
      row[_row].addElement(_column, _index);
      col[_column].addElement(_row, _index);
    } else
      set(_index, _entry);
  }

  inline void addToElement(long _row, long _column, double _entry) {
    double *_index = getIndex(_row, _column);
    if (_index == NULL) {
      if (!_entry) return;
      _index = insert(_entry);
      row[_row].addElement(_column, _index);
      col[_column].addElement(_row, _index);
    } else
      set(_index, _entry + (*_index) );
  }

  inline void divideColElement(SparseMatrixCol *_col, long index, double factor)
  {
    double *_index = _col->getIndex(index);
    set(_index, (*_index)/factor);
  }

  inline double getColElement(SparseMatrixCol *_col, long _index) const {
    double *index = _col->getIndex(_index);
    return (*index);
  }

  inline double getRowElement(SparseMatrixRow *_row, long _index) const {
    double *index = _row->getIndex(_index);
    return (*index);
  }

  void print();
};


/*
class SparseMatrix {
protected:
  long rows, cols, elements, calls;
  FastDynaArray<SparseMatrixCol> col;
  AVLTree *tree;
  Store *store;
public:
  SparseMatrix(long r, long c);
  ~SparseMatrix();
  void statistics();
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getCalls() const { return calls; }
  inline long getElements() const { return elements; }
  inline SparseMatrixCol &getCol(long index) { return col(index); }

  inline double *insert(double _entry) {
    #ifdef DEBUG
    cout << "insert: searching for " << _entry << endl;
    #endif
    calls++;
    double *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
      elements++;
      cout << "store now has " << elements  << " elements" << endl;
    }
    return _index;
  }

  inline void addElement(long _row, long _column, double _entry) {
    double *_index = insert(_entry);    
    col[_column].addElement(_row, _index);
  }

  inline double getElement(long _row, long _column) const {
    for (long n=0; n<col(_column).getEntries(); n++)
      if (col(_column).getRow(n) == _row) return *(col(_column).getIndex(n));
    return 0.0;     
  }

  inline double getColElement(SparseMatrixCol *_col, long _index) const {
    return *(_col->getIndex(_index));
  }

  int invert();

  void transMultiply(Vector<double> &x, Vector <double> &result) const;
  void preMultiply(Vector<double> &x, Vector<double> &result) const;
  void calcResidual(Vector<double> &x, Vector <double> &result) const;
  void SORIteration(Vector<double> &x, double omega) const;
  void safeSORIteration(Vector<double> &x, double omega) const;
  void twinSORIteration(Vector<double> &x1, double omega1, Vector<double>
&x2, double omega2) const;
  void calcResidualNorm(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const;
  void calcResidualNormAI(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const;
  void twinCalcResidualNorm(Vector<double> &x1, double &x1Norm, double &x1InfNorm,
Vector<double> &x2, double &x2Norm, double &x2InfNorm, Vector<double> &last,
double &x1DiffNorm, double &x2DiffNorm) const;

  void print();
};
*/











