#ifndef __STATE_HPP
#define __STATE_HPP

class StateHashTable {
protected:
  long accesses, maxRowLength;
  FastDynaArray<StateHashTableRow> rows; // ### set to HASH_LENGTH
public:
  StateHashTable();
  ~StateHashTable();

  inline int registerState(long primary, long secondary, long target) {
#ifdef VERBOSE
    cout << "StateHashTable (" << this << ") registering state (" << primary << "," << secondary << ") -> ("<< target << ")\n";
#endif
    rows[primary].add(secondary, target);
    if (rows(primary).getEntries() > maxRowLength) maxRowLength = rows(primary).getEntries();
    return 1;
  }

  inline int searchState(long primary, long secondary, long &id) {
    accesses++;
    return rows[primary].search(secondary, id);
  }

  inline long getAccesses() { return accesses; }
  inline int getMaxRowLength() { return maxRowLength; }
  void statistics();
};

class QueueEntry {
protected:
  FullState *item;
  QueueEntry *next;
public:
  QueueEntry(FullState *current) {
    item = current;
    next = NULL;
  }
  inline FullState *getItem() { return item; }
  inline QueueEntry *getNext() { return next; }
  inline void setItem(FullState *_item) { item = _item; }
  inline void setNext(QueueEntry *_next) { next = _next; }
};

class StateQueue {
protected:
  QueueEntry *front, *rear;
public:
  StateQueue() { front = rear = NULL; }
  ~StateQueue() {}
  inline FullState *top() { return front->getItem(); }
  inline void push(FullState *current) {
    QueueEntry *item = new QueueEntry(current);
    if (!front) {
      front = rear = item;
      return;
    }
    rear->setNext(item);
    rear = item;
  }
  inline FullState *pop() {
#ifdef SAFE
    assert(front);
#endif
    FullState *value = front->getItem();
    QueueEntry *next = front->getNext();
    delete front;
    front = next;
    return value;
  }
};

class Generator {
protected:
  FILE *sf, *mf;
  char *buffer[BUFFERS];
  void *core;

  StateQueue *tangibleStack;
  ClusterStack *vanishStack;
  StateHashTable *hashTable;
  Timer *timer;
  Information *info;

  ReportStyle repStyle;
  long unexplored, maxUnexplored, maxStates, tangible, repInterval;
  double currentRatio, maxCPUTime;
  int maxChildren, watchClock;

  int openFile(char *filename, FILE *&fp, char *buff);
  int closeFile(FILE *fp);

  void terminate(char *fmt, ...);
  void destroy();

  void pushStart(FullState *);
  long pushTangible(FullState *current) {
#ifdef VERBOSE
    cout << "Generator (" << this << "): push state (" << tangible << ") keys (" << current->getPrimaryKey() << "," << current->getSecondaryKey() << ")\n";
    current->output();
#endif

#ifdef SAFE  
    assert (current->getType()==TANGIBLE);
#endif
#ifdef DEBUG
    long id;
    assert(hashTable->searchState(current->getPrimaryKey(), current->getSecondaryKey(),id)==0);
#endif

    hashTable->registerState(current->getPrimaryKey(), current->getSecondaryKey(), tangible);
    tangible++;
    tangibleStack->push(current);

#ifdef VERBOSE
    cout << "Generator Push: " << tangible << " states and " << unexplored+1 << " unexplored markings\n";
#endif

    unexplored++;
#ifndef SILENT
    if (repInterval && !(tangible % repInterval) && repStyle != NONE) {
      currentRatio = (double) unexplored/tangible;
      cout << tangible << " states generated (" << unexplored << " unexplored) r = " << currentRatio << endl;
    }
    if (!(tangible % 10000)) {
      if (watchClock && (timer->observe() > maxCPUTime))
        terminate("maximum cpu time (%lf) exceeded", maxCPUTime);
      if (maxStates && tangible > maxStates)
        terminate("maximum state number (%ld) exceeded", maxStates);
    }
#endif

    if (unexplored>maxUnexplored) maxUnexplored = unexplored;
    return tangible - 1;
  }

  inline int pushVanishing(FullState *current, double weight) {
    if (weight > EPS*10) {
      vanishStack->push(current, weight);
      return 1;
    } else 
      return 0;
  }

  inline FullState *popTangible() {
#ifdef VERBOSE
    cout << "Generator (" << this << ") pop state (" << unexplored-1 << ")\n";
#endif
#ifdef SAFE
    assert(unexplored>=0);
#endif
    unexplored--;
    FullState *returnValue = tangibleStack->pop();
#ifdef SAFE
    assert(returnValue);
    assert(returnValue->getMagic() == 1973);
#endif
#ifdef VERBOSE
    returnValue->output();
#endif
    return returnValue;
  }   

  inline void popVanishing(FullState* &target, double &weight) {
#ifdef VERBOSE
    cout << "Generator (" << this << ") pop vanishing state (" << vanishing-1 << ")\n";
#endif
    vanishStack->pop(target, weight);
#ifdef SAFE
    assert(target);
    assert(target->getMagic() == 1973);
    assert(target->getType()==VANISHING);
#endif
#ifdef VERBOSE
    target->output();
#endif
  }   

  void statistics();
  void deadlock(FullState *current) {
    cout << "warning: no transitions enabled (deadlock) in state:" << endl;
    current->output();
  }
public:
  Generator(char*);
  ~Generator();
  int explore(FullState *, char*);
};

#endif


